pub mod format;
mod printers;

use crate::formatter::format::Format;
use rust_embed::RustEmbed;
use serde::Deserialize;
use std::error::Error;
use std::fmt::{Display, Formatter};
use std::fs::File;
use std::io::Read;
use std::path::PathBuf;
use std::str;

#[derive(Debug)]
#[allow(clippy::enum_variant_names)]
pub enum FormatConfigError {
    IOError(std::io::Error),
    TomlError {
        file_name: String,
        err: toml::de::Error,
    },
    Utf8Error(std::str::Utf8Error),
}

impl Error for FormatConfigError {}

impl From<std::io::Error> for FormatConfigError {
    fn from(e: std::io::Error) -> Self {
        Self::IOError(e)
    }
}

impl From<std::str::Utf8Error> for FormatConfigError {
    fn from(e: std::str::Utf8Error) -> Self {
        Self::Utf8Error(e)
    }
}

impl Display for FormatConfigError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let err_msg = match self {
            FormatConfigError::IOError(e) => e.to_string(),
            FormatConfigError::TomlError { file_name, err } => {
                format!("Error parsing {}: {}", file_name, err)
            }
            FormatConfigError::Utf8Error(e) => e.to_string(),
        };

        write!(f, "Format Config Error: {}", err_msg)
    }
}

#[derive(RustEmbed)]
#[folder = "formats/"]
#[include = "*.toml"]
struct BuiltInFormats;

#[derive(Debug, Deserialize, Clone, Default)]
pub struct FormatConfig {
    pub formats: Vec<Format>,
}

impl FormatConfig {
    pub fn new(config_path: &Option<PathBuf>) -> Result<Self, FormatConfigError> {
        let mut config_str = String::new();

        let mut config: FormatConfig = if let Some(config_path) = config_path {
            let mut config = File::open(config_path)?;
            config.read_to_string(&mut config_str)?;

            toml::from_str(&config_str).map_err(|e| FormatConfigError::TomlError {
                file_name: config_path.to_str().unwrap().to_string(),
                err: e,
            })?
        } else {
            FormatConfig::default()
        };

        for format_file_path in BuiltInFormats::iter() {
            if format_file_path.ends_with("md") {
                continue;
            }

            let format_file = BuiltInFormats::get(&format_file_path).unwrap();

            let config_str = str::from_utf8(&format_file.data).unwrap();

            let mut built_in: FormatConfig =
                toml::from_str(config_str).map_err(|e| FormatConfigError::TomlError {
                    file_name: format_file_path.to_string(),
                    err: e,
                })?;

            config.formats.append(&mut built_in.formats);
        }

        Ok(config)
    }
}
